/*
 * Copyright (c) 2014-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Wed Mar  5 2014
 */
#include "acoptitrack.h"

#include <err.h>
#include <stdio.h>

#include "optitrack_c_types.h"
#include "codels.h"


/* --- opti_create_ports --------------------------------------------------- */

genom_event
opti_create_ports(const struct natnet_descr *descr,
                  const optitrack_ids_pubdata_s *pubdata,
                  const optitrack_bodies *bodies,
                  const optitrack_markers *markers,
                  genom_context self)
{
  const struct natnet_bodydescr *body;
  genom_event e;
  uint32_t i;

  for(i = 0, body = descr->body; i < descr->nbodies; i++, body++) {
    if (pubdata->bodies) {
      if (!bodies->data(body->safename, self)) {
        e = bodies->open(body->safename, self);
        if (e != genom_ok) {
          warnx("failed to export body '%s'", body->safename);
          return e;
        }
      }
    }

    if (pubdata->markers) {
      if (!markers->data(body->safename, self)) {
        e = markers->open(body->safename, self);
        if (e != genom_ok) {
          warnx("failed to export markers '%s'", body->safename);
          return e;
        }
      }
    }
  }

  if (pubdata->anon && !markers->data("anonymous", self)) {
    e = markers->open("anonymous", self);
    if (e != genom_ok) {
      warnx("failed to export anonymous markers");
      return e;
    }
  }

  return genom_ok;
}


/* --- opti_update_ports --------------------------------------------------- */

genom_event
opti_update_ports(const struct natnet_frame *f,
                  optitrack_server_s *server,
                  optitrack_bodies_info_s *binfo,
                  const optitrack_ids_pubdata_s *pubdata,
                  const optitrack_bodies *bodies,
                  const optitrack_markers *markers,
                  FILE *log, genom_context self)
{
  const struct natnet_body *body;
  genom_event e;
  uint32_t i, j;

  for(i=0, body=f->body; i<f->nbodies; i++, body++) {
    if (body->descr == NULL) {
      warnx("body %d has gone", f->body[i].id);
      binfo->descr = NULL; /* refetch descriptions */
      continue;
    }

    if (pubdata->bodies) {
      or_pose_estimator_state *state =
        bodies->data(body->descr->safename, self);
      if (!state) {
        warnx("body %s port does not exist", body->descr->safename);
        return optitrack_e_publish(self);
      }

      /* frame timestamp */
      state->ts.sec = f->timestamp.tv_sec;
      state->ts.nsec = f->timestamp.tv_usec * 1000;
      state->intrinsic = false;
 
      state->pos._present = (body->pos && body->error > 0.);
      state->pos_cov._present = state->pos._present;
      state->vel._present = 0;
      state->vel_cov._present = 0;
      state->acc._present = 0;
      state->acc_cov._present = 0;

      if (state->pos._present) {
        double qw, qx, qy, qz;
        double s;

        /* remap optitrack coordinates into something more usual (Z up) */
        state->pos._value.x = body->pos->z;
        state->pos._value.y = body->pos->x;
        state->pos._value.z = body->pos->y;

        qw = state->pos._value.qw = body->pos->qw;
        qx = state->pos._value.qx = body->pos->qz;
        qy = state->pos._value.qy = body->pos->qx;
        qz = state->pos._value.qz = body->pos->qy;

        /* provide fake but hopefully realistic covariance */
        for(j=0; j<sizeof(state->pos_cov._value.cov)/sizeof(double); j++)
          state->pos_cov._value.cov[j] = 0.;

        s = 10. * body->error * body->error;
        state->pos_cov._value.cov[0] = s;
        state->pos_cov._value.cov[2] = s;
        state->pos_cov._value.cov[5] = s;

        s = 100. * body->error * body->error;
        state->pos_cov._value.cov[9]  = s * (1 - qw*qw);
        state->pos_cov._value.cov[13] = s * -qw*qx;
        state->pos_cov._value.cov[14] = s * (1 - qx*qx);
        state->pos_cov._value.cov[18] = s * qw*qy;
        state->pos_cov._value.cov[19] = s * -qx*qy;
        state->pos_cov._value.cov[20] = s * (1 - qy*qy);
        state->pos_cov._value.cov[24] = s * -qw*qz;
        state->pos_cov._value.cov[25] = s * -qx*qz;
        state->pos_cov._value.cov[26] = s * -qy*qz;
        state->pos_cov._value.cov[27] = s * (1 - qz*qz);
      }

      e = bodies->write(body->descr->safename, self);
      if (e != genom_ok) {
        warnx("body %s cannot be published", body->descr->safename);
        return e;
      }

      if (log && state->pos._present) {
        fprintf(log, "%s %d %d   %g %g %g   %g %g %g %g   %g\n",
                body->descr->safename,
                state->ts.sec, state->ts.nsec,
                state->pos._value.x,
                state->pos._value.y,
                state->pos._value.z,
                state->pos._value.qw,
                state->pos._value.qx,
                state->pos._value.qy,
                state->pos._value.qz,
                body->error);
      }
    }

    if (pubdata->markers) {
      uint32_t i;
      optitrack_marker_set *state =
        markers->data(body->descr->safename, self);
      if (!state) {
        warnx("markers %s port does not exist", body->descr->safename);
        return optitrack_e_publish(self);
      }

      if (!body->pos || body->error == 0.) /* not tracked */ continue;

      /* frame timestamp */
      state->ts.sec = f->timestamp.tv_sec;
      state->ts.nsec = f->timestamp.tv_usec * 1000;
 
      /* markers sequence */
      if (state->markers._maximum != body->mset.n)
        if (genom_sequence_reserve(&state->markers, body->mset.n))
          return optitrack_e_nomem(self);

      for(i = 0; i < body->mset.n; i++) {
        /* remap optitrack coordinates into something more usual (Z up) */
        state->markers._buffer[i].x = body->mset.pos[i].z;
        state->markers._buffer[i].y = body->mset.pos[i].x;
        state->markers._buffer[i].z = body->mset.pos[i].y;
      }
      state->markers._length = body->mset.n;

      e = markers->write(body->descr->safename, self);
      if (e != genom_ok) {
        warnx("markers %s cannot be published", body->descr->safename);
        return e;
      }

      if (log) {
        for(i = 0; i < body->mset.n; i++) {
          fprintf(log, "%s:%d %d %d   %g %g %g\n",
                  body->descr->safename, i,
                  state->ts.sec, state->ts.nsec,
                  state->markers._buffer[i].x,
                  state->markers._buffer[i].y,
                  state->markers._buffer[i].z);
        }
      }
    }
  }

  if (pubdata->anon) {
    uint32_t i;
    optitrack_marker_set *state = markers->data("anonymous", self);
    if (!state) {
      warnx("anonymous markers cannot be published");
      return optitrack_e_publish(self);
    }

    /* frame timestamp */
    state->ts.sec = f->timestamp.tv_sec;
    state->ts.nsec = f->timestamp.tv_usec * 1000;
 
    /* markers sequence */
    if (state->markers._maximum < f->aset.n)
      if (genom_sequence_reserve(&state->markers, f->aset.n))
        return optitrack_e_nomem(self);

    for(i = 0; i < f->aset.n; i++) {
      /* remap optitrack coordinates into something more usual (Z up) */
      state->markers._buffer[i].x = f->aset.pos[i].z;
      state->markers._buffer[i].y = f->aset.pos[i].x;
      state->markers._buffer[i].z = f->aset.pos[i].y;
    }
    state->markers._length = f->aset.n;

    e = markers->write("anonymous", self);
    if (e != genom_ok) {
      warnx("anonymous markers cannot be published");
      return e;
    }

    if (log) {
      for(i = 0; i < f->aset.n; i++) {
        fprintf(log, "anonymous:%d %d %d   %g %g %g\n",
                i,
                state->ts.sec, state->ts.nsec,
                state->markers._buffer[i].x,
                state->markers._buffer[i].y,
                state->markers._buffer[i].z);
      }
    }
  }

  return genom_ok;
}
